import base64
import dataclasses
import re
from dataclasses import InitVar
from functools import cached_property
from typing import Optional, Iterable, Any, Final

import clingo
import clingo.ast
import typeguard
from clingo.ast import ComparisonOperator, Location
from dumbo_utils.primitives import PrivateKey
from dumbo_utils.validation import validate

from dumbo_asp import utils
from dumbo_asp.primitives.atoms import SymbolicAtom, GroundAtom
from dumbo_asp.primitives.models import Model
from dumbo_asp.primitives.parsers import Parser
from dumbo_asp.primitives.predicates import Predicate
from dumbo_asp.primitives.terms import SymbolicTerm
from dumbo_asp.utils import uuid


@typeguard.typechecked
@dataclasses.dataclass(frozen=True)
class SymbolicRule:
    __value: clingo.ast.AST
    __parsed_string: Optional[str]
    disabled: bool

    key: InitVar[PrivateKey]
    __key = PrivateKey()

    def __post_init__(self, key: PrivateKey):
        self.__key.validate(key)
        validate("type", self.__value.ast_type, equals=clingo.ast.ASTType.Rule)

    @staticmethod
    def parse(string: str, disabled: bool = False) -> "SymbolicRule":
        program = Parser.parse_program(string)
        validate("one rule", program, length=1,
                 help_msg=f"Unexpected sequence of {len(program)} rules in {utils.one_line(string)}")
        return SymbolicRule(program[0], string, disabled=disabled, key=SymbolicRule.__key)

    @staticmethod
    def of(value: clingo.ast.AST, disabled: bool = False) -> "SymbolicRule":
        validate("value", value.ast_type == clingo.ast.ASTType.Rule, equals=True)
        return SymbolicRule(value, None, disabled=disabled, key=SymbolicRule.__key)

    def __str__(self):
        res = self.__parsed_string or str(self.__value)
        return f"%* {res} *%" if self.disabled else res

    def transform(self, transformer: clingo.ast.Transformer) -> Any:
        transformer(self.__value)

    @property
    def is_fact(self) -> bool:
        return len(self.__value.body) == 0 and self.is_normal_rule

    @property
    def is_normal_rule(self) -> bool:
        return self.__value.head.ast_type == clingo.ast.ASTType.Literal and \
            self.__value.head.sign == clingo.ast.Sign.NoSign

    @property
    def is_choice_rule(self) -> bool:
        return self.__value.head.ast_type == clingo.ast.ASTType.Aggregate

    @property
    def is_disjunctive_rule(self) -> bool:
        return self.__value.head.ast_type == clingo.ast.ASTType.Disjunction

    @property
    def is_constraint(self) -> bool:
        return self.head_atom == SymbolicAtom.of_false()

    @property
    def head_atom(self) -> SymbolicAtom:
        if ("atom" in self.__value.head.keys()) and ("value" in self.__value.head.atom.keys()):
            validate("#false", self.__value.head.atom.value, equals=0)
            return SymbolicAtom.of_false()
        return SymbolicAtom.of(self.__value.head.atom.symbol)

    @staticmethod
    def __compute_choice_bounds(choice):
        left, right = 0, "unbounded"
        if choice.left_guard is not None:
            validate("left guard", choice.left_guard.comparison != ComparisonOperator.NotEqual, equals=True)
            if choice.left_guard.comparison == ComparisonOperator.LessThan:
                left = f"{choice.left_guard.term} + 1"
            elif choice.left_guard.comparison == ComparisonOperator.LessEqual:
                left = f"{choice.left_guard.term}"
            elif choice.left_guard.comparison == ComparisonOperator.GreaterThan:
                right = f"{choice.left_guard.term} - 1"
            elif choice.left_guard.comparison == ComparisonOperator.GreaterEqual:
                right = f"{choice.left_guard.term}"
            elif choice.left_guard.comparison == ComparisonOperator.Equal:
                left = f"{choice.left_guard.term}"
                right = f"{choice.left_guard.term}"
            else:
                raise ValueError("Choice with != are not supported.")
        if choice.right_guard is not None:
            validate("right guard", choice.right_guard.comparison, is_in=[ComparisonOperator.LessThan,
                                                                          ComparisonOperator.LessEqual])
            if choice.right_guard.comparison == ComparisonOperator.LessThan:
                right = f"{choice.right_guard.term} + 1"
            elif choice.right_guard.comparison == ComparisonOperator.LessEqual:
                right = f"{choice.right_guard.term}"
        return left, right

    @property
    def choice_lower_bound(self) -> str:
        validate("choice rule", self.is_choice_rule, equals=True)
        return self.__compute_choice_bounds(self.__value.head)[0]

    @property
    def choice_upper_bound(self) -> str:
        validate("choice rule", self.is_choice_rule, equals=True)
        return self.__compute_choice_bounds(self.__value.head)[1]

    @property
    def positive_body_literals(self) -> tuple[SymbolicAtom, ...]:
        return tuple(SymbolicAtom.of(literal.atom.symbol)
                     for literal in self.__value.body
                     if literal.sign == clingo.ast.Sign.NoSign and "symbol" in literal.atom.keys())

    @property
    def negative_body_literals(self) -> tuple[SymbolicAtom, ...]:
        return tuple(SymbolicAtom.of(literal.atom.symbol)
                     for literal in self.__value.body
                     if literal.sign == clingo.ast.Sign.Negation and "symbol" in literal.atom.keys())

    def serialize(self, *, base64_encode: bool = True) -> tuple[GroundAtom, ...]:
        def b64(s):
            return f'"{base64.b64encode(str(s).encode()).decode()}"' if base64_encode else \
                str(clingo.String(str(s)))
        rule = b64(self)
        res = [f'rule({rule})']
        if self.is_normal_rule:
            if not self.is_constraint:
                res.append(f"head({rule}, {b64(self.head_atom)})")
        elif self.is_choice_rule:
            lb, ub = self.__compute_choice_bounds(self.__value.head)
            res.append(f"choice({rule}, {lb}, {ub})")
            for atom in self.__value.head.elements:
                assert not atom.condition  # extend to conditional
                res.append(f"head({rule}, {b64(atom)})")
        elif self.is_disjunctive_rule:
            for atom in self.__value.head.elements:
                assert not atom.condition  # extend to conditional
                res.append(f"head({rule}, {b64(atom)})")
        else:
            assert False
        for literal in self.__value.body:
            if "atom" not in literal.keys():
                assert False  # extend?
            if literal.sign == clingo.ast.Sign.NoSign:
                predicate = "pos_body"
            elif literal.sign == clingo.ast.Sign.Negation:
                predicate = "neg_body"
            else:
                assert False  # extend

            if literal.atom.ast_type == clingo.ast.ASTType.Comparison:
                if Model.empty().compute_substitutions(
                    arguments="",
                    number_of_arguments=0,
                    conjunctive_query=str(literal.atom)
                ):
                    if predicate == "pos_body":
                        continue
                else:
                    if predicate == "neg_body":
                        continue
            res.append(f'{predicate}({rule}, {b64(literal.atom)})')
        return tuple(GroundAtom.parse(atom) for atom in res)

    @cached_property
    def head_variables(self) -> tuple[str, ...]:
        res = set()

        class Transformer(clingo.ast.Transformer):
            def visit_Variable(self, node):
                res.add(str(node))
                return node

        Transformer().visit(self.__value.head)
        return tuple(sorted(res))

    @cached_property
    def body_variables(self) -> tuple[str, ...]:
        res = set()

        class Transformer(clingo.ast.Transformer):
            def visit_Variable(self, node):
                res.add(str(node))
                return node

        Transformer().visit_sequence(self.__value.body)
        return tuple(sorted(res))

    @cached_property
    def global_safe_variables(self) -> tuple[str, ...]:
        res = set()

        class Transformer(clingo.ast.Transformer):
            def visit_Literal(self, node):
                if node.sign == clingo.ast.Sign.NoSign:
                    self.visit_children(node)

            def visit_ConditionalLiteral(self, node):
                # a conditional literal cannot bound new variables
                # (this is not the case for "existential" variables, which are not covered at the moment)
                pass

            def visit_BodyAggregate(self, node):
                for guard in [node.left_guard, node.right_guard]:
                    if guard is not None and guard.comparison == clingo.ast.ComparisonOperator.Equal:
                        self.visit(guard.term)

            def visit_Variable(self, node):
                if node.name != '_':
                    res.add(node.name)
                return node

        Transformer().visit_sequence(self.__value.body)
        return tuple(sorted(res))

    @cached_property
    def predicates(self) -> tuple[Predicate, ...]:
        res = set()

        class Transformer(clingo.ast.Transformer):
            def visit_Function(self, node):
                res.add((node.name, len(node.arguments)))
                return node

            def visit_Literal(self, node):
                if "symbol" in node.atom.keys():
                    res.add((node.atom.symbol.name, len(node.atom.symbol.arguments)))
                elif "elements" in node.atom.keys():
                    for element in node.atom.elements:
                        self.visit(element.update(terms=[]))
                return node

        Transformer().visit(self.__value)
        return tuple(Predicate.parse(*pred) for pred in res)

    def disable(self) -> "SymbolicRule":
        return SymbolicRule(self.__value, self.__parsed_string, True, key=self.__key)

    def with_extended_body(self, atom: SymbolicAtom, sign: clingo.ast.Sign = clingo.ast.Sign.NoSign) -> "SymbolicRule":
        literal = f"{atom}" if sign == clingo.ast.Sign.NoSign else \
            f"not {atom}" if sign == clingo.ast.Sign.Negation else \
            f"not not {atom}"
        if self.__parsed_string is None:
            string = str(self)
            line = 1
            column = len(string) - 1
        else:
            string = self.__parsed_string
            line = self.__value.location.end.line
            column = self.__value.location.end.column - 1
        new_rule = utils.insert_in_parsed_string(
            f"; {literal}" if len(self.__value.body) > 0 else f" :- {literal}", string, line, column
        )
        return self.parse(new_rule, self.disabled)

    def body_as_string(self, separator: str = "; ") -> str:
        return separator.join(str(x) for x in self.__value.body)

    def apply_variable_substitution(self, **kwargs: SymbolicTerm) -> "SymbolicRule":
        class Transformer(clingo.ast.Transformer):
            def visit_Variable(self, node):
                if str(node) not in kwargs.keys():
                    return node
                return kwargs[str(node)].make_copy_of_value()

        return self.of(Transformer().visit(self.__value), self.disabled)

    def apply_term_substitution(self, **kwargs: SymbolicTerm) -> "SymbolicRule":
        class Transformer(clingo.ast.Transformer):
            def visit_SymbolicTerm(self, node):
                if str(node) not in kwargs.keys():
                    return node
                return kwargs[str(node)].make_copy_of_value()

        return self.of(Transformer().visit(self.__value), self.disabled)

    def apply_predicate_renaming(self, **kwargs: Predicate) -> "SymbolicRule":
        class Transformer(clingo.ast.Transformer):
            def visit_Function(self, node):
                return node.update(name=kwargs[node.name].name) if node.name in kwargs.keys() else node

        return self.of(Transformer().visit(self.__value), self.disabled)

    def __expand_global_safe_variables(
            self,
            *,
            variables: Iterable[str],
            herbrand_base: Model,
            expand_also_local_variables=False,
    ) -> tuple["SymbolicRule", ...]:
        the_variables: Final = set(var for var in variables if var in self.global_safe_variables)
        validate("variables", set(variables), equals=the_variables)
        substitutions = herbrand_base.compute_substitutions(
            arguments=','.join(the_variables),
            number_of_arguments=len(the_variables),
            conjunctive_query=self.body_as_string(),
        ) if the_variables else ([],)

        prefix = uuid()
        suffix = uuid()

        class Transformer(clingo.ast.Transformer):
            def __init__(self):
                super().__init__()
                self.possibly_has_local_variables = False
                self.locations = []

            def visit_Variable(self, node):
                if node.name in the_variables:
                    self.locations.append((node.location, prefix + node.name + suffix))
                return node

            # def visit_BodyAggregateElement(self, node):  NOT SUPPORTED AT THE MOMENT
            def visit_ConditionalLiteral(self, node):
                self.possibly_has_local_variables = True
                self.visit_children(node)
                return node

        transformer = Transformer()
        transformer.visit(self.__value)
        fmt = self.__parsed_string or str(self.__value)
        for location, replacement in reversed(transformer.locations):
            fmt = utils.replace_in_parsed_string(fmt, location, replacement)

        pattern = f"{prefix}({'|'.join(var for var in the_variables)}){suffix}"
        var_to_index = {var: index for index, var in enumerate(the_variables)}

        def apply(substitution):
            return re.sub(pattern, lambda m: substitution[var_to_index[m.group(1)]], fmt)

        if expand_also_local_variables and transformer.possibly_has_local_variables:
            return tuple(
                SymbolicRule.parse(apply([str(s) for s in substitution]), self.disabled)
                .__expand_local_variables(herbrand_base=herbrand_base)
                for substitution in substitutions
            )
        else:
            return tuple(
                SymbolicRule.parse(apply([str(s) for s in substitution]), self.disabled)
                for substitution in substitutions
            )

    def __expand_local_variables(self, *, herbrand_base: Model) -> "SymbolicRule":
        class Transformer(clingo.ast.Transformer):
            def __init__(self):
                super().__init__()
                self.substitutions = []

            # def visit_BodyAggregateElement(self, node):  NOT SUPPORTED AT THE MOMENT
            def visit_ConditionalLiteral(self, node):
                substitutions = herbrand_base.compute_substitutions(
                    arguments=','.join(str(arg) for arg in node.literal.atom.symbol.arguments),
                    number_of_arguments=len(node.literal.atom.symbol.arguments),
                    conjunctive_query=f"{', '.join(str(condition) for condition in node.condition)}",
                )
                self.substitutions.append(
                    (
                        Location(
                            begin=node.location.begin,
                            end=node.condition[-1].location.end if node.condition else node.location.end
                        ),
                        [
                            (
                                "not " if node.literal.sign == clingo.ast.Sign.Negation else
                                "not not " if node.literal.sign == clingo.ast.Sign.DoubleNegation
                                else ""
                            ) +
                            f"{node.literal.atom.symbol.name}({','.join(str(arg) for arg in arguments)})" if arguments
                            else f"{node.literal.atom.symbol.name}"
                            for arguments in substitutions
                        ]
                    )
                )
                return node

        transformer = Transformer()
        transformer.visit(self.__value)
        rule = self.__parsed_string or str(self.__value)
        for location, atoms in reversed(transformer.substitutions):
            rule = utils.replace_in_parsed_string(rule, location, '; '.join(atoms))
        return SymbolicRule.parse(rule, disabled=self.disabled)

    def expand_global_safe_variables(
            self,
            *,
            variables: Iterable[str],
            herbrand_base: Model,
    ) -> tuple["SymbolicRule", ...]:
        return self.__expand_global_safe_variables(variables=variables, herbrand_base=herbrand_base,
                                                   expand_also_local_variables=False)

    def expand_global_and_local_variables(self, *, herbrand_base: Model) -> tuple["SymbolicRule", ...]:
        return self.__expand_global_safe_variables(variables=self.global_safe_variables, herbrand_base=herbrand_base,
                                                   expand_also_local_variables=True)

    def match(self, *pattern: SymbolicAtom) -> bool:
        class Transformer(clingo.ast.Transformer):
            def visit_SymbolicAtom(self, node):
                atom = SymbolicAtom.of(node.symbol)
                if atom.match(*pattern):
                    Transformer.matched = True
                return node
        Transformer.matched = False

        Transformer().visit(self.__value)
        return Transformer.matched

    def to_zero_simplification_version(self, *, compact=False) -> "SymbolicRule":
        if compact:
            atom = Predicate.false().name
        else:
            rule_vars_as_strings = ','.join(f'"{var}"' for var in self.global_safe_variables)
            rule_id = f'("{base64.b64encode(str(self).encode()).decode()}", ' \
                      f'({rule_vars_as_strings}{"," if len(rule_vars_as_strings) == 1 else ""}))'
            rule_vars = ','.join(self.global_safe_variables)

            atom = f'{Predicate.false().name}({rule_id}, ' \
                   f'({rule_vars}{"," if len(rule_vars) == 1 else ""}))'

        if self.is_choice_rule:
            if self.__value.head.elements:
                _, line, column = self.__value.head.elements[0].location.begin
                return SymbolicRule.parse(utils.insert_in_parsed_string(f"{atom}; ", str(self), line, column))
            s = str(self)
            index = 0
            while True:
                if s[index] == '%':
                    index += 1
                    if s[index] == '*':
                        index += 1
                        while s[index] != '*' or s[index + 1] != '%':
                            index += 1
                        index += 1
                    else:
                        index += 1
                        while s[index] != '\n':
                            index += 1
                if s[index] == '{':
                    break
                index += 1
            return SymbolicRule.parse(s[:index + 1] + f"{atom}" + s[index + 1:])
        if self.is_constraint:
            return SymbolicRule.parse(f'{atom}\n{self}')
        return SymbolicRule.parse(f'{atom} |\n{self}')



from uuid import uuid4
from pathlib import Path
from typing import Final

import clingo.ast
import typeguard

PROJECT_ROOT: Final = Path(__file__).parent.parent
NEW_LINE_SYMBOL: Final = '⏎'


@typeguard.typechecked
def extract_parsed_string(string: str, location: clingo.ast.Location) -> str:
    lines = string.split('\n')
    res = []
    if location.begin.line == location.end.line:
        res.append(lines[location.begin.line - 1][location.begin.column - 1:location.end.column - 1])
    else:
        res.append(lines[location.begin.line - 1][location.begin.column - 1:])
        res.extend(lines[location.begin.line:location.end.line - 1])
        res.append(lines[location.end.line - 1][:location.end.column - 1])
    return '\n'.join(line for line in res)


@typeguard.typechecked
def insert_in_parsed_string(addendum: str, string: str, line: int, column: int) -> str:
    lines = string.split('\n')
    lines[line - 1] = lines[line - 1][:column - 1] + addendum + lines[line - 1][column - 1:]
    return '\n'.join(lines)


@typeguard.typechecked
def replace_in_parsed_string(string: str, location: clingo.ast.Location, new_content: str) -> str:
    lines = string.split('\n')
    lines[location.begin.line - 1] = lines[location.begin.line - 1][:location.begin.column - 1] + \
                                     new_content + \
                                     lines[location.end.line - 1][location.end.column - 1:]
    lines = lines[:location.begin.line] + lines[location.end.line:]
    return '\n'.join(line for line in lines)


@typeguard.typechecked
def one_line(string: str) -> str:
    return NEW_LINE_SYMBOL.join(string.split('\n'))


def uuid() -> str:
    return str(uuid4()).replace('-', '_')



import dataclasses
import functools
from dataclasses import InitVar
from typing import Optional

import clingo
import clingo.ast
import typeguard
from dumbo_utils.primitives import PrivateKey
from dumbo_utils.validation import validate

from dumbo_asp.primitives.parsers import Parser


@functools.total_ordering
@typeguard.typechecked
@dataclasses.dataclass(frozen=True)
class Predicate:
    name: str
    arity: Optional[int]

    key: InitVar[PrivateKey]
    __key = PrivateKey()

    MAX_ARITY = 999

    def __post_init__(self, key: PrivateKey):
        self.__key.validate(key)

    @staticmethod
    def parse(name: str, arity: Optional[int] = None) -> "Predicate":
        split = name.split('/', maxsplit=1)
        if len(split) == 2:
            validate("arity not given", arity is None, equals=True, help_msg="The arity is given already in the name")
            name, arity = split[0], int(split[1])

        term = Parser.parse_ground_term(name)
        validate("name", term.type, equals=clingo.SymbolType.Function)
        validate("name", term.arguments, length=0)
        validate("name", term.negative, equals=False)
        if arity is not None:
            validate("arity", arity, min_value=0, max_value=Predicate.MAX_ARITY)
        return Predicate(
            name=term.name,
            arity=arity,
            key=Predicate.__key,
        )

    @staticmethod
    def of(term: clingo.Symbol) -> "Predicate":
        return Predicate(
            name=term.name,
            arity=len(term.arguments),
            key=Predicate.__key,
        )

    def drop_arity(self) -> "Predicate":
        return Predicate(
            name=self.name,
            arity=None,
            key=Predicate.__key,
        )

    def with_arity(self, arity: int) -> "Predicate":
        validate("arity", arity, min_value=0, max_value=Predicate.MAX_ARITY)
        return Predicate(
            name=self.name,
            arity=arity,
            key=Predicate.__key,
        )

    def match(self, other: "Predicate") -> bool:
        if self.name != other.name:
            return False
        if self.arity is None or other.arity is None:
            return True
        return self.arity == other.arity

    def __lt__(self, other: "Predicate"):
        if self.name < other.name:
            return True
        if self.name > other.name:
            return False

        if self.arity is None:
            return False
        if other.arity is None:
            return True

        return self.arity < other.arity

    @staticmethod
    def false() -> "Predicate":
        return Predicate.of(clingo.Function("__false__")).drop_arity()



import copy
import dataclasses
import functools
from dataclasses import InitVar
from functools import cached_property
from typing import Optional, Final

import clingo
import clingo.ast
import typeguard
from dumbo_utils.primitives import PrivateKey
from dumbo_utils.validation import validate

from dumbo_asp import utils
from dumbo_asp.primitives.parsers import Parser
from dumbo_asp.primitives.predicates import Predicate
from dumbo_asp.primitives.terms import SymbolicTerm


@functools.total_ordering
@typeguard.typechecked
@dataclasses.dataclass(frozen=True)
class GroundAtom:
    value: clingo.Symbol

    def __post_init__(self):
        validate("atom format", self.value.type, equals=clingo.SymbolType.Function,
                 help_msg="An atom must have a predicate name")

    @staticmethod
    def parse(string: str) -> "GroundAtom":
        return GroundAtom(Parser.parse_ground_term(string))

    @cached_property
    def predicate(self) -> Predicate:
        return Predicate.of(self.value)

    @property
    def predicate_name(self) -> str:
        return self.predicate.name

    @property
    def predicate_arity(self) -> int:
        return self.predicate.arity

    @cached_property
    def arguments(self) -> tuple[clingo.Symbol, ...]:
        return tuple(self.value.arguments)

    @property
    def strongly_negated(self) -> bool:
        return self.value.negative

    def __str__(self):
        return str(self.value)

    def __lt__(self, other: "GroundAtom"):
        if self.predicate < other.predicate:
            return True
        if self.predicate == other.predicate:
            for index, argument in enumerate(self.arguments):
                other_argument = other.arguments[index]
                if argument.type < other_argument.type:
                    return True
                if argument.type > other_argument.type:
                    return False
                if argument.type == clingo.SymbolType.Number:
                    if argument < other_argument:
                        return True
                    if argument > other_argument:
                        return False
                else:
                    s1, s2 = str(argument), str(other_argument)
                    if s1 < s2:
                        return True
                    if s1 > s2:
                        return False
        return False


@typeguard.typechecked
@dataclasses.dataclass(frozen=True)
class SymbolicAtom:
    __value: clingo.ast.AST
    __parsed_string: Optional[str]

    key: InitVar[PrivateKey]
    __key = PrivateKey()

    def __post_init__(self, key: PrivateKey):
        self.__key.validate(key)
        validate("type", self.__value.ast_type,
                 is_in=[clingo.ast.ASTType.SymbolicAtom, clingo.ast.ASTType.Function,
                        clingo.ast.ASTType.BooleanConstant])

    @staticmethod
    def of_false() -> "SymbolicAtom":
        return SymbolicAtom.parse("#false")

    @staticmethod
    def of_ground_atom(atom: GroundAtom) -> "SymbolicAtom":
        return SymbolicAtom.parse(str(atom))

    @staticmethod
    def parse(string: str) -> "SymbolicAtom":
        rule: Final = f":- {string}."
        try:
            program = Parser.parse_program(rule)
        except Parser.Error as error:
            raise error.drop(first=3, last=1)

        validate("one rule", program, length=1,
                 help_msg=f"Unexpected sequence of {len(program)} rules in {utils.one_line(string)}")
        validate("one atom", program[0].body, length=1,
                 help_msg=f"Unexpected conjunction of {len(program[0].body)} atoms in {utils.one_line(string)}")
        literal = program[0].body[0]
        validate("positive", literal.sign, equals=clingo.ast.Sign.NoSign,
                 help_msg=f"Unexpected default negation in {utils.one_line(string)}")
        if "value" in literal.atom.keys():
            validate("#false", literal.atom.value, equals=0)
            atom = SymbolicAtom.parse("foo").__value.update(name="#false")
        else:
            atom = literal.atom.symbol
        return SymbolicAtom(atom, utils.extract_parsed_string(rule, literal.location), key=SymbolicAtom.__key)

    @staticmethod
    def of(value: clingo.ast.AST) -> "SymbolicAtom":
        validate("value", value.ast_type, is_in=[
            clingo.ast.ASTType.Function
        ])
        return SymbolicAtom(value, None, key=SymbolicAtom.__key)

    def __str__(self):
        return self.__parsed_string or str(self.__value)

    def make_copy_of_value(self) -> clingo.ast.AST:
        return copy.deepcopy(self.__value)

    @cached_property
    def predicate(self) -> Predicate:
        return Predicate.parse(self.__value.name, len(self.__value.arguments))

    @property
    def predicate_name(self) -> str:
        if self.__value.name == '#false':
            return self.__value.name
        return self.predicate.name

    @property
    def predicate_arity(self) -> int:
        return self.predicate.arity

    @cached_property
    def arguments(self) -> tuple[SymbolicTerm, ...]:
        return tuple(SymbolicTerm.parse(str(argument)) for argument in self.__value.arguments)

    @property
    def strongly_negated(self) -> bool:
        return self.value.negative

    def match(self, *pattern: "SymbolicAtom") -> bool:
        for a_pattern in pattern:
            if self.predicate == a_pattern.predicate and \
                    all(argument.match(a_pattern.arguments[index]) for index, argument in enumerate(self.arguments)):
                return True
        return False



import copy
import dataclasses
from dataclasses import InitVar
from functools import cached_property
from typing import Optional, Final

import clingo
import clingo.ast
import typeguard
from dumbo_utils.primitives import PrivateKey
from dumbo_utils.validation import validate

from dumbo_asp import utils
from dumbo_asp.primitives.parsers import Parser


@typeguard.typechecked
@dataclasses.dataclass(frozen=True)
class SymbolicTerm:
    __value: clingo.ast.AST
    __parsed_string: Optional[str]

    key: InitVar[PrivateKey]
    __key = PrivateKey()

    def __post_init__(self, key: PrivateKey):
        self.__key.validate(key)
        validate("type", self.__value.ast_type, is_in=[clingo.ast.ASTType.SymbolicTerm, clingo.ast.ASTType.Function,
                                                       clingo.ast.ASTType.Variable])

    @staticmethod
    def parse(string: str) -> "SymbolicTerm":
        rule: Final = f":- a({string})."
        try:
            program = Parser.parse_program(rule)
        except Parser.Error as error:
            raise error.drop(first=5, last=2)

        validate("one rule", program, length=1,
                 help_msg=f"Unexpected sequence of {len(program)} rules in {utils.one_line(string)}")
        validate("one atom", program[0].body, length=1,
                 help_msg=f"Unexpected conjunction of {len(program[0].body)} atoms in {utils.one_line(string)}")
        atom = program[0].body[0].atom.symbol
        validate("arity", atom.arguments, length=1,
                 help_msg=f"Unexpected sequence of {len(atom.arguments)} terms in {utils.one_line(string)}")
        return SymbolicTerm(atom.arguments[0], utils.extract_parsed_string(rule, atom.arguments[0].location),
                            key=SymbolicTerm.__key)

    @staticmethod
    def of_int(value: int) -> "SymbolicTerm":
        return SymbolicTerm.parse(str(value))

    @staticmethod
    def of_string(value: str) -> "SymbolicTerm":
        return SymbolicTerm.parse(f'"{value}"')

    def __str__(self):
        return self.__parsed_string or str(self.__value)

    def is_int(self) -> bool:
        return self.__value.ast_type == clingo.ast.ASTType.SymbolicTerm and \
            self.__value.symbol.type == clingo.SymbolType.Number

    def is_string(self) -> bool:
        return self.__value.ast_type == clingo.ast.ASTType.SymbolicTerm and \
            self.__value.symbol.type == clingo.SymbolType.String

    def is_function(self) -> bool:
        return self.__value.ast_type == clingo.ast.ASTType.Function or \
            self.__value.ast_type == clingo.ast.ASTType.SymbolicTerm and \
            self.__value.symbol.type == clingo.SymbolType.Function

    def is_variable(self) -> bool:
        return self.__value.ast_type == clingo.ast.ASTType.Variable

    def int_value(self) -> int:
        return self.__value.symbol.number

    def string_value(self) -> str:
        return self.__value.symbol.string

    @property
    def function_name(self) -> str:
        return self.__value.name if self.__value.ast_type == clingo.ast.ASTType.Function else self.__value.symbol.name

    @property
    def function_arity(self) -> int:
        return len(self.__value.arguments if self.__value.ast_type == clingo.ast.ASTType.Function else
                   self.__value.symbol.arguments)

    @cached_property
    def arguments(self) -> tuple["SymbolicTerm", ...]:
        return tuple(SymbolicTerm.parse(str(argument)) for argument in self.__value.arguments) \
            if "arguments" in self.__value.keys() else ()

    def make_copy_of_value(self) -> clingo.ast.AST:
        return copy.deepcopy(self.__value)

    def match(self, pattern: "SymbolicTerm") -> bool:
        if pattern.is_variable() or self.is_variable():
            return True
        if pattern.is_function():
            return self.is_function() and pattern.function_name == self.function_name and \
                pattern.function_arity == self.function_arity and \
                all(argument.match(pattern.arguments[index]) for index, argument in enumerate(self.arguments))
        return pattern == self




import base64
import json
import subprocess
import webbrowser
import zlib
from pathlib import Path
from typing import Final, Iterable

import clingo
import igraph
import typeguard
from clingox.reify import reify_program
from dumbo_utils.primitives import PositiveIntegerOrUnbounded
from dumbo_utils.url import compress_object_for_url
from dumbo_utils.validation import validate

from dumbo_asp import utils
from dumbo_asp.primitives.atoms import GroundAtom, SymbolicAtom
from dumbo_asp.primitives.models import Model
from dumbo_asp.primitives.programs import SymbolicProgram
from dumbo_asp.primitives.rules import SymbolicRule
from dumbo_asp.primitives.terms import SymbolicTerm


@typeguard.typechecked
def compute_minimal_unsatisfiable_subsets(
        program: SymbolicProgram,
        up_to: PositiveIntegerOrUnbounded = PositiveIntegerOrUnbounded.of(1),
        *,
        over_the_ground_program: bool = False,
        clingo_path: Path = Path("clingo"),
        wasp: Path = Path("wasp"),
) -> list[SymbolicProgram]:
    predicate: Final = f"__mus__"
    if over_the_ground_program:
        rules = [
            SymbolicRule.parse(f"__constant{predicate}({';'.join(str(term) for term in program.herbrand_universe)}).")
        ]
        for index, rule in enumerate(program, start=1):
            terms = ','.join([str(index), *rule.global_safe_variables])
            rules.append(rule.with_extended_body(SymbolicAtom.parse(f"{predicate}({terms})")))
            variables = '; '.join(f"__constant{predicate}({variable})" for variable in rule.global_safe_variables)
            rules.append(SymbolicRule.parse(f"{{{predicate}({terms})}} :- {variables}."))
        mus_program = SymbolicProgram.of(rules)
    else:
        mus_program = SymbolicProgram.of(
            *(rule.with_extended_body(SymbolicAtom.parse(f"{predicate}({index})"))
              for index, rule in enumerate(program, start=1)),
            SymbolicRule.parse(
                f"{{{predicate}(1..{len(program)})}}."
            ),
        )
    # print(mus_program)
    res = subprocess.run(
        ["bash", "-c",
         f"{clingo_path} --output=smodels | {wasp} --silent --mus={predicate} -n {up_to if up_to.is_int else 0}"],
        input=str(mus_program).encode(),
        capture_output=True,
    )
    validate("exit code", res.returncode, equals=0, help_msg="Computation failed")
    lines = res.stdout.decode().split('\n')
    muses = [Model.of_atoms(line.split()[2:]) for line in lines if line]
    if not over_the_ground_program:
        return [SymbolicProgram.of(program[atom.arguments[0].number - 1] for atom in mus) for mus in muses]
    res = []
    for mus in muses:
        rules = []
        for atom in mus:
            rule = program[atom.arguments[0].number - 1]
            rules.append(rule.apply_variable_substitution(**{
                variable: SymbolicTerm.parse(str(atom.arguments[index]))
                for index, variable in enumerate(rule.global_safe_variables, start=1)
            }))
        res.append(SymbolicProgram.of(rules))
    return res


@typeguard.typechecked
def enumerate_models(
        program: SymbolicProgram, *,
        true_atoms: Iterable[GroundAtom] = (),
        false_atoms: Iterable[GroundAtom] = (),
        unknown_atoms: Iterable[GroundAtom] = (),
        up_to: int = 0,
) -> tuple[Model, ...]:
    """
    Enumerate models of the program that are compatible with the partial assignment.
    Note that the program may be simplified by clingo, so you may want to specify some unknown atoms to prevent
    such simplifications.
    """
    validate("up_to", up_to, min_value=0)

    the_program = Model.of_atoms(
        reify_program(
            Model.of_atoms(true_atoms).as_facts +
            '\n'.join(f":- {atom}." for atom in false_atoms) +
            Model.of_atoms(unknown_atoms).as_choice_rules +
            str(program)
        )
    ).as_facts + META_MODELS

    return __collect_models(the_program, [f"{up_to}"])


@typeguard.typechecked
def enumerate_counter_models(
        program: SymbolicProgram,
        model: Model,
        *,
        up_to: int = 0,
) -> tuple[Model, ...]:
    validate("up_to", up_to, min_value=0)

    the_program = Model.of_atoms(
        reify_program(
            '\n'.join(f"#external {atom}." for atom in model) +
            str(program)
        )
    ).as_facts + META_COUNTER_MODELS + '\n'.join(f"true(L) :- output({atom},B), literal_tuple(B,L)." for atom in model)

    return __collect_models(the_program, [f"{up_to}"])


@typeguard.typechecked
def validate_in_all_models(
        program: SymbolicProgram, *,
        true_atoms: Iterable[GroundAtom] = (),
        false_atoms: Iterable[GroundAtom] = (),
        unknown_atoms: Iterable[GroundAtom] = (),
) -> None:
    the_program = Model.of_atoms(
        reify_program(
            Model.of_atoms(true_atoms, false_atoms, unknown_atoms).as_choice_rules +
            str(program)
        )
    ).as_facts + META_MODELS

    def check(mode: bool, atoms):
        consequences = set(
            at for at in __collect_models(the_program, ["--enum-mode=cautious" if mode else "--enum-mode=brave"])[-1]
        )
        for atom in atoms:
            validate(f"{mode} atom", atom in consequences, equals=mode,
                     help_msg=f"Atom {atom} was expected to be {str(mode).lower()} in all models")

    check(True, true_atoms)
    check(False, false_atoms)


@typeguard.typechecked
def validate_in_all_models_of_the_reduct(
        program: SymbolicProgram, *,
        model: Model,
        true_atoms: Iterable[GroundAtom] = (),
) -> None:
    the_program = Model.of_atoms(
        reify_program(
            '\n'.join(f"#external {atom}." for atom in model) +
            str(program)
        )
    ).as_facts + META_REDUCT_MODELS + '\n'.join(f"true(L) :- output({atom},B), literal_tuple(B,L)." for atom in model)
    consequences = set(
        at for at in __collect_models(the_program, ["--enum-mode=cautious"])[-1]
    )
    for atom in true_atoms:
        validate(f"True atom", atom in consequences, equals=True,
                 help_msg=f"Atom {atom} was expected to be true in all models")


@typeguard.typechecked
def validate_cannot_be_true_in_any_stable_model(
        program: SymbolicProgram,
        atom: GroundAtom,
        *,
        unknown_atoms: Iterable[GroundAtom] = (),
        local_prefix: str = "__",
) -> None:
    false_in_all_models = False
    try:
        validate_in_all_models(program=program, false_atoms=(atom,), unknown_atoms=unknown_atoms)
        false_in_all_models = True
    except ValueError:
        pass
    if false_in_all_models:
        return

    models = enumerate_models(program, true_atoms=(atom,), unknown_atoms=unknown_atoms)
    for model in models:
        the_program = SymbolicProgram.of(
            *program,
            (SymbolicRule.parse(f"{at}.") for at in model if not at.predicate_name.startswith(local_prefix))
        )
        validate("has counter model", enumerate_counter_models(the_program, model, up_to=1), length=1)


@typeguard.typechecked
def validate_cannot_be_extended_to_stable_model(
        program: SymbolicProgram,
        *,
        true_atoms: Iterable[GroundAtom] = (),
        false_atoms: Iterable[GroundAtom] = (),
        unknown_atoms: Iterable[GroundAtom] = (),
        local_prefix: str = "__",
) -> None:
    false_in_all_models = False
    try:
        fail = f"__fail_{utils.uuid()}"
        validate_in_all_models(program=SymbolicProgram.of(
            (
                *program,
                SymbolicRule.parse(f"\n{fail} :- " + '; '.join(
                    [f"{atom}" for atom in true_atoms] + [f"not {atom}" for atom in false_atoms]
                ) + '.')
            )
        ), false_atoms=(GroundAtom.parse(fail),), unknown_atoms=(*true_atoms, *false_atoms, *unknown_atoms))
        false_in_all_models = True
    except ValueError:
        pass
    if false_in_all_models:
        return

    models = enumerate_models(program, true_atoms=true_atoms, false_atoms=false_atoms, unknown_atoms=unknown_atoms)
    for model in models:
        the_program = SymbolicProgram.of(
            *program,
            (SymbolicRule.parse(f"{at}.") for at in model if not at.predicate_name.startswith(local_prefix))
        )
        validate("has counter model", enumerate_counter_models(the_program, model, up_to=1), length=1)


def __collect_models(program: str, options: list[str]) -> tuple[Model, ...]:
    control = clingo.Control(options)
    control.add(program)
    control.ground([("base", [])])
    res = []

    def collect(model):
        res.append(Model.of_atoms(model.symbols(shown=True)))

    control.solve(on_model=collect)
    return tuple(res)


@typeguard.typechecked
def pack_asp_chef_url(recipe: str, the_input: str | Model | Iterable[Model]) -> str:
    if type(the_input) == Model:
        the_input = the_input.as_facts
    elif type(the_input) != str:
        the_input = '§'.join(model.as_facts for model in the_input)
    url = recipe.replace("/#", "/open#", 1)
    url = url.replace(r"#.*;", "#", 1)
    url = url.replace("#", "#" + compress_object_for_url({"input": the_input}, suffix="") + ";", 1)
    return url


@typeguard.typechecked
def open_graph_in_xasp_navigator(graph_model: Model):
    reason_map: Final = {
        "true": {
            "support": "support",
            "constraint": "required true to falsify body",
            "last support": "required true to satisfy body of last supporting rule",
        },
        "false": {
            "lack of support": "lack of support",
            "choice": "required false to satisfy choice rule upper bound",
            "constraint": "required false to falsify body",
            "last support": "required false to satisfy body of last supporting rule",
        },
    }

    graph = igraph.Graph(directed=True)

    atom_to_rule = {}
    for node in graph_model.filter(when=lambda atom: atom.predicate_name == "node"):
        name = node.arguments[0].string
        value = node.arguments[1].name
        reason = node.arguments[2].arguments[0].name.replace('_', ' ')
        atom_to_rule[name] = node.arguments[2].arguments[1].string if len(node.arguments[2].arguments) == 2 else "" # fix (the label possibly come from the link)
        graph.add_vertex(name, label=f"{name}\n{reason_map[value][reason]}")

    for link in graph_model.filter(when=lambda atom: atom.predicate_name == "link"):
        source = link.arguments[0].string
        target = link.arguments[1].string
        label = atom_to_rule[source]
        graph.add_edge(source, target, label=label)

    # layout = graph.layout_sugiyama()
    layout = graph.layout_reingold_tilford()
    res = {
        "nodes": [
            {
                "id": index,
                "label": node.attributes()["label"],
                "x": layout.coords[index][0],
                "y": layout.coords[index][1],
            }
            for index, node in enumerate(graph.vs)
        ],
        "links": [
            {
                "source": link.tuple[0],
                "target": link.tuple[1],
                "label": link.attributes()["label"],
            }
            for link in graph.es
        ],
    }
    url = "https://xasp-navigator.netlify.app/#"
    # url = "http://localhost:5173/#"
    json_dump = json.dumps(res, separators=(',', ':')).encode()
    url += base64.b64encode(zlib.compress(json_dump)).decode() + '%21'
    webbrowser.open(url, new=0, autoraise=True)


META_MODELS = """
atom( A ) :- atom_tuple(_,A).
atom(|L|) :-          literal_tuple(_,L).
atom(|L|) :- weighted_literal_tuple(_,L).

{ hold(A) : atom(A) }.

conjunction(B) :- literal_tuple(B),
        hold(L) : literal_tuple(B, L), L > 0;
    not hold(L) : literal_tuple(B,-L), L > 0.

body(normal(B)) :- rule(_,normal(B)), conjunction(B).
body(sum(B,G))  :- rule(_,sum(B,G)),
    #sum { W,L :     hold(L), weighted_literal_tuple(B, L,W), L > 0 ;
           W,L : not hold(L), weighted_literal_tuple(B,-L,W), L > 0 } >= G.

  hold(A) : atom_tuple(H,A)   :- rule(disjunction(H),B), body(B).
{ hold(A) : atom_tuple(H,A) } :- rule(choice(H),B), body(B).

#show.
#show T : output(T,B), conjunction(B).

% avoid warnings
atom_tuple(0,0) :- #false.
conjunction(0) :- #false.
literal_tuple(0) :- #false.
literal_tuple(0,0) :- #false.
weighted_literal_tuple(0,0) :- #false.
weighted_literal_tuple(0,0,0) :- #false.
rule(0,0) :- #false.
"""

META_COUNTER_MODELS = """
atom( A ) :- atom_tuple(_,A).
atom(|L|) :-          literal_tuple(_,L).
atom(|L|) :- weighted_literal_tuple(_,L).

{hold(A)} :- atom(A), true(A).
:- hold(A) : true(A).

conjunction(B) :- literal_tuple(B),
        hold(L) : literal_tuple(B, L), L > 0;
    not true(L) : literal_tuple(B,-L), L > 0.

body(normal(B)) :- rule(_,normal(B)), conjunction(B).
body(sum(B,G))  :- rule(_,sum(B,G)),
    #sum { W,L :     hold(L), weighted_literal_tuple(B, L,W), L > 0 ;
           W,L : not true(L), weighted_literal_tuple(B,-L,W), L > 0 } >= G.

  hold(A) : atom_tuple(H,A)   :- rule(disjunction(H),B), body(B).
{ hold(A) : atom_tuple(H,A) } :- rule(     choice(H),B), body(B).

#show.
#show T : output(T,B), conjunction(B).

% avoid warnings
atom_tuple(0,0) :- #false.
conjunction(0) :- #false.
literal_tuple(0) :- #false.
literal_tuple(0,0) :- #false.
weighted_literal_tuple(0,0) :- #false.
weighted_literal_tuple(0,0,0) :- #false.
rule(0,0) :- #false.
"""

META_REDUCT_MODELS = """
atom( A ) :- atom_tuple(_,A).
atom(|L|) :-          literal_tuple(_,L).
atom(|L|) :- weighted_literal_tuple(_,L).

{hold(A)} :- atom(A), true(A).
:- not hold(A), true(A).

conjunction(B) :- literal_tuple(B),
        hold(L) : literal_tuple(B, L), L > 0;
    not true(L) : literal_tuple(B,-L), L > 0.

body(normal(B)) :- rule(_,normal(B)), conjunction(B).
body(sum(B,G))  :- rule(_,sum(B,G)),
    #sum { W,L :     hold(L), weighted_literal_tuple(B, L,W), L > 0 ;
           W,L : not true(L), weighted_literal_tuple(B,-L,W), L > 0 } >= G.

  hold(A) : atom_tuple(H,A)   :- rule(disjunction(H),B), body(B).
{ hold(A) : atom_tuple(H,A) } :- rule(     choice(H),B), body(B).

#show.
#show T : output(T,B), conjunction(B).

% avoid warnings
atom_tuple(0,0) :- #false.
conjunction(0) :- #false.
literal_tuple(0) :- #false.
literal_tuple(0,0) :- #false.
weighted_literal_tuple(0,0) :- #false.
weighted_literal_tuple(0,0,0) :- #false.
rule(0,0) :- #false.
"""

META_HT_MODELS = """
#const option=1.

atom( A ) :- atom_tuple(_,A).
atom(|L|) :-          literal_tuple(_,L).
atom(|L|) :- weighted_literal_tuple(_,L).

model(h). model(t).

{ hold(A,h) } :- atom(A),    option = 1.
{ hold(A,t) } :- atom(A).
:- hold(L,h), not hold(L,t).

conjunction(B,M) :- model(M), literal_tuple(B),
        hold(L,M) : literal_tuple(B, L), L > 0;
    not hold(L,t) : literal_tuple(B,-L), L > 0.

body(normal(B),M) :- rule(_,normal(B)), conjunction(B,M).
body(sum(B,G),M)  :- model(M), rule(_,sum(B,G)),
    #sum { W,L :     hold(L,M), weighted_literal_tuple(B, L,W), L > 0 ;
           W,L : not hold(L,t), weighted_literal_tuple(B,-L,W), L > 0 } >= G.

               hold(A,M) :  atom_tuple(H,A)   :- rule(disjunction(H),B), body(B,M).
hold(A,M); not hold(A,t) :- atom_tuple(H,A),     rule(     choice(H),B), body(B,M).

#show.
#show (T,M) : output(T,B), conjunction(B,M).

% avoid warnings
atom_tuple(0,0) :- #false.
conjunction(0) :- #false.
literal_tuple(0) :- #false.
literal_tuple(0,0) :- #false.
weighted_literal_tuple(0,0) :- #false.
weighted_literal_tuple(0,0,0) :- #false.
rule(0,0) :- #false.
"""


import base64
import dataclasses
from collections import defaultdict
from dataclasses import InitVar
from functools import cached_property, cache
from typing import Optional, Iterable, Dict

import clingo
import clingo.ast
import typeguard
from dumbo_utils.primitives import PrivateKey
from dumbo_utils.validation import validate

from dumbo_asp import utils
from dumbo_asp.primitives.atoms import GroundAtom, SymbolicAtom
from dumbo_asp.primitives.models import Model
from dumbo_asp.primitives.parsers import Parser
from dumbo_asp.primitives.predicates import Predicate
from dumbo_asp.primitives.rules import SymbolicRule
from dumbo_asp.primitives.terms import SymbolicTerm
from dumbo_asp.utils import uuid


@typeguard.typechecked
@dataclasses.dataclass(frozen=True)
class SymbolicProgram:
    __rules: tuple[SymbolicRule, ...]
    __parsed_string: Optional[str]

    key: InitVar[PrivateKey]
    __key = PrivateKey()

    def __post_init__(self, key: PrivateKey):
        self.__key.validate(key)

    @staticmethod
    def of(*args: SymbolicRule | Iterable[SymbolicRule]) -> "SymbolicProgram":
        rules = []
        for arg in args:
            if type(arg) == SymbolicRule:
                rules.append(arg)
            else:
                rules.extend(arg)
        return SymbolicProgram(tuple(rules), None, key=SymbolicProgram.__key)

    @staticmethod
    def parse(string: str) -> "SymbolicProgram":
        rules = tuple(SymbolicRule.parse(utils.extract_parsed_string(string, rule.location))
                      for rule in Parser.parse_program(string))
        return SymbolicProgram(rules, string, key=SymbolicProgram.__key)

    def __str__(self):
        return '\n'.join(str(rule) for rule in self.__rules) if self.__parsed_string is None else self.__parsed_string

    def __len__(self):
        return len(self.__rules)

    def __getitem__(self, item: int):
        return self.__rules[item]

    def __iter__(self):
        for rule in self.__rules:
            yield rule

    @cached_property
    def herbrand_universe(self) -> set[SymbolicTerm]:
        res = set()

        def get_arguments(term):
            for argument in term.arguments:
                if argument.type == clingo.SymbolType.Function and argument.arguments:
                    get_arguments(argument)
                else:
                    res.add(SymbolicTerm.parse(str(argument)))

        for atom in self.herbrand_base:
            get_arguments(atom)
        return res

    @cached_property
    def herbrand_base(self) -> Model:
        control = clingo.Control()
        control.add(str(self))
        control.ground([("base", [])])
        return Model.of_atoms(atom.symbol for atom in control.symbolic_atoms)

    @cached_property
    def herbrand_base_without_false_predicate(self) -> Model:
        return self.herbrand_base.drop(Predicate.false())

    @cached_property
    def herbrand_base_false_predicate_only(self) -> Model:
        return self.herbrand_base.filter(when=lambda at: at.predicate == Predicate.false().with_arity(2))

    @cached_property
    def rules_grouped_by_false_predicate(self):
        atoms = self.herbrand_base_false_predicate_only
        res = defaultdict(list)
        variables = {}
        for atom in atoms:
            key = base64.b64decode(atom.arguments[0].arguments[0].string.encode()).decode()
            res[key].append(atom.arguments[1])
            variables[key] = {arg.string: index for index, arg in enumerate(atom.arguments[0].arguments[1].arguments)}
        return res, variables

    def serialize(self, *, base64_encode: bool = True) -> tuple[GroundAtom, ...]:
        res = []
        for rule in self:
            res.extend(rule.serialize(base64_encode=base64_encode))
        return tuple(dict.fromkeys(res))

    @cached_property
    def predicates(self) -> tuple[Predicate, ...]:
        res = set()
        for rule in self:
            res.update(rule.predicates)
        return tuple(res)

    @cache
    def process_constants(self) -> "SymbolicProgram":
        rules = []
        constants = {}
        for rule in self:
            if rule.is_fact:
                head_atom = rule.head_atom
                if head_atom.predicate_name == "__const__":
                    validate("arity", head_atom.predicate_arity, equals=2, help_msg="Error in defining constant")
                    name, value = head_atom.arguments
                    constants[str(name)] = value
                    rules.append(rule.disable())
                    continue
            rules.append(rule.apply_term_substitution(**constants))

        return SymbolicProgram.of(rules)

    @cache
    def process_with_statements(self) -> "SymbolicProgram":
        rules = []
        statements_queue = []
        for rule in self:
            if rule.is_fact:
                head_atom = rule.head_atom
                if head_atom.predicate_name == "__with__":
                    statements_queue.append(tuple(SymbolicAtom.parse(str(argument))
                                                  for argument in head_atom.arguments))
                    rules.append(rule.disable())
                    continue
                if head_atom.predicate_name == "__end_with__":
                    validate("no arguments", head_atom.arguments, length=0)
                    statements_queue.pop()
                    rules.append(rule.disable())
                    continue
            for statement in statements_queue:
                for literal in statement:
                    rule = rule.with_extended_body(literal)
            rules.append(rule)
        validate("all __with__ are terminated", statements_queue, length=0,
                 help_msg=f"{len(statements_queue)} unterminated __with__ statements")

        return SymbolicProgram.of(rules)

    def apply_predicate_renaming(self, **kwargs: Predicate) -> "SymbolicProgram":
        return SymbolicProgram.of(rule.apply_predicate_renaming(**kwargs) for rule in self)

    def expand_global_safe_variables(self, *, rule: SymbolicRule, variables: Iterable[str],
                                     herbrand_base: Optional[Model] = None) -> "SymbolicProgram":
        rules = []
        for __rule in self.__rules:
            if rule != __rule:
                rules.append(__rule)
            else:
                rules.extend(__rule.expand_global_safe_variables(
                    variables=variables,
                    herbrand_base=self.herbrand_base_without_false_predicate if herbrand_base is None else herbrand_base
                ))
        return SymbolicProgram.of(rules)

    def expand_global_safe_variables_in_rules(
            self,
            rules_to_variables: Dict[SymbolicRule, Iterable[str]],
            herbrand_base: Optional[Model] = None,
    ) -> "SymbolicProgram":
        rules = []
        for __rule in self.__rules:
            if __rule in rules_to_variables.keys():
                rules.extend(__rule.expand_global_safe_variables(
                    variables=rules_to_variables[__rule],
                    herbrand_base=self.herbrand_base_without_false_predicate
                    if herbrand_base is None else herbrand_base,
                ))
            else:
                rules.append(__rule)
        return SymbolicProgram.of(rules)

    def expand_global_and_local_variables(self, *, expand_also_disabled_rules: bool = False,
                                          herbrand_base: Optional[Model] = None) -> "SymbolicProgram":
        rules = []
        for rule in self.__rules:
            if not rule.disabled or expand_also_disabled_rules:
                rules.extend(
                    rule.expand_global_and_local_variables(
                        herbrand_base=self.herbrand_base_without_false_predicate
                        if herbrand_base is None else herbrand_base
                    )
                )
            else:
                rules.append(rule)
        return SymbolicProgram.of(rules)

    def move_up(self, *pattern: SymbolicAtom) -> "SymbolicProgram":
        def key(rule: SymbolicRule):
            return 0 if rule.match(*pattern) else 1
        return SymbolicProgram.of(sorted([rule for rule in self.__rules], key=key))

    def query_herbrand_base(self, query_head_arguments: str, query_body: str,
                            aux_program: Optional["SymbolicProgram"] = None) -> tuple[SymbolicAtom, ...]:
        predicate = f"__query_{uuid()}__"
        program = SymbolicProgram.parse(f"{self.herbrand_base.as_facts}\n"
                                        f"{predicate}({query_head_arguments}) :- {query_body}.")
        if aux_program is not None:
            program = SymbolicProgram.of(*program, *aux_program)
        model = Model.of_program(str(program)).filter(lambda atom: atom.predicate.name == predicate)
        return tuple(SymbolicAtom.of_ground_atom(atom) for atom in model)

    def to_zero_simplification_version(self, *, extra_atoms: Iterable[GroundAtom] = (), 
                                       compact=False) -> "SymbolicProgram":
        false_predicate = Predicate.false().name
        return SymbolicProgram.of(
            [rule.to_zero_simplification_version(compact=compact) for rule in self],
            SymbolicRule.parse(' | '.join(str(atom) for atom in extra_atoms) + f" :- {false_predicate}.")
            if extra_atoms else [],
            SymbolicRule.parse(f"{{{false_predicate}}}."),
            SymbolicRule.parse(f":- {false_predicate}.") if compact else
            SymbolicRule.parse(f":- #count{{0 : {false_predicate}; "
                               f"RuleID, Substitution "
                               f": {false_predicate}(RuleID, Substitution)}} > 0."),
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dumbo_asp', 'dumbo_asp.primitives']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.3.0,<10.0.0',
 'cairocffi>=1.6.0,<2.0.0',
 'clingo>=5.6.2,<6.0.0',
 'clingox>=1.2.0,<2.0.0',
 'dateutils>=0.6.12,<0.7.0',
 'distlib>=0.3.7,<0.4.0',
 'dumbo-utils>=0.1.8,<0.2.0',
 'igraph>=0.11.3,<0.12.0']

setup_kwargs = {
    'name': 'dumbo-asp',
    'version': '0.1.4',
    'description': 'Utilities for Answer Set Programming',
    'long_description': 'None',
    'author': 'Mario Alviano',
    'author_email': 'mario.alviano@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
